import copy
import importlib
import json
import math
import os
import sys
from collections import namedtuple
from types import ModuleType
from typing import Any

import pytest

import betterproto2
from tests.inputs import config as test_input_config
from tests.mocks import MockChannel
from tests.util import find_module, get_directories, get_test_case_json_data, inputs_path

# Force pure-python implementation instead of C++, otherwise imports
# break things because we can't properly reset the symbol database.
os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"

from google.protobuf.json_format import Parse


class TestCases:
    def __init__(
        self,
        path,
        services: set[str],
        xfail: set[str],
    ):
        all = set(get_directories(path)) - {"__pycache__"}
        messages = {test for test in all - services if get_test_case_json_data(test)}

        unknown_xfail_tests = xfail - all
        if unknown_xfail_tests:
            raise Exception(f"Unknown test(s) in config.py: {unknown_xfail_tests}")

        self.services = self.apply_xfail_marks(services, xfail)
        self.messages = self.apply_xfail_marks(messages, xfail)

    @staticmethod
    def apply_xfail_marks(test_set: set[str], xfail: set[str]):
        return [pytest.param(test, marks=pytest.mark.xfail) if test in xfail else test for test in test_set]


test_cases = TestCases(
    path=inputs_path,
    services=test_input_config.services,
    xfail=test_input_config.xfail,
)

plugin_output_package = "tests.output_betterproto"
reference_output_package = "tests.output_reference"

TestData = namedtuple("TestData", ["plugin_module", "reference_module", "json_data"])


def module_has_entry_point(module: ModuleType):
    return any(hasattr(module, attr) for attr in ["Test", "TestStub"])


def list_replace_nans(items: list) -> list[Any]:
    """Replace float("nan") in a list with the string "NaN"

    Parameters
    ----------
    items : List
            List to update

    Returns
    -------
    List[Any]
        Updated list
    """
    result = []
    for item in items:
        if isinstance(item, list):
            result.append(list_replace_nans(item))
        elif isinstance(item, dict):
            result.append(dict_replace_nans(item))
        elif isinstance(item, float) and math.isnan(item):
            result.append(betterproto2.NAN)
    return result


def dict_replace_nans(input_dict: dict[Any, Any]) -> dict[Any, Any]:
    """Replace float("nan") in a dictionary with the string "NaN"

    Parameters
    ----------
    input_dict : Dict[Any, Any]
            Dictionary to update

    Returns
    -------
    Dict[Any, Any]
        Updated dictionary
    """
    result = {}
    for key, value in input_dict.items():
        if isinstance(value, dict):
            value = dict_replace_nans(value)
        elif isinstance(value, list):
            value = list_replace_nans(value)
        elif isinstance(value, float) and math.isnan(value):
            value = betterproto2.NAN
        result[key] = value
    return result


@pytest.fixture
def reset_sys_path():
    original = copy.deepcopy(sys.path)
    yield
    sys.path = original


@pytest.fixture
def test_data(request, reset_sys_path):
    test_case_name = request.param

    reference_module_root = os.path.join(*reference_output_package.split("."), test_case_name)
    sys.path.append(reference_module_root)

    plugin_module = importlib.import_module(f"{plugin_output_package}.{test_case_name}")

    plugin_module_entry_point = find_module(plugin_module, module_has_entry_point)

    if not plugin_module_entry_point:
        raise Exception(
            f"Test case {repr(test_case_name)} has no entry point. "
            "Please add a proto message or service called Test and recompile."
        )

    yield (
        TestData(
            plugin_module=plugin_module_entry_point,
            reference_module=lambda: importlib.import_module(
                f"{reference_output_package}.{test_case_name}.{test_case_name}_pb2"
            ),
            json_data=get_test_case_json_data(test_case_name),
        )
    )


@pytest.mark.parametrize("test_data", test_cases.messages, indirect=True)
def test_message_can_instantiated(test_data: TestData) -> None:
    plugin_module, *_ = test_data
    plugin_module.Test()


@pytest.mark.parametrize("test_data", test_cases.messages, indirect=True)
def test_message_equality(test_data: TestData) -> None:
    plugin_module, *_ = test_data
    message1 = plugin_module.Test()
    message2 = plugin_module.Test()
    assert message1 == message2


@pytest.mark.parametrize("test_data", test_cases.messages, indirect=True)
def test_message_json(test_data: TestData) -> None:
    plugin_module, _, json_data = test_data

    for sample in json_data:
        if sample.belongs_to(test_input_config.non_symmetrical_json):
            continue

        message: betterproto2.Message = plugin_module.Test.from_json(sample.json)
        message_json = message.to_json(indent=0)

        assert dict_replace_nans(json.loads(message_json)) == dict_replace_nans(json.loads(sample.json))


@pytest.mark.parametrize("test_data", test_cases.services, indirect=True)
def test_service_can_be_instantiated(test_data: TestData) -> None:
    test_data.plugin_module.TestStub(MockChannel())


@pytest.mark.parametrize("test_data", test_cases.messages, indirect=True)
def test_binary_compatibility(test_data: TestData) -> None:
    plugin_module, reference_module, json_data = test_data

    # TODO fix and delete
    if "map" in plugin_module.__file__.replace("\\", "/").split("/"):
        pytest.skip("Skipping this test for now.")

    for sample in json_data:
        reference_instance = Parse(sample.json, reference_module().Test())
        reference_binary_output = reference_instance.SerializeToString()

        plugin_instance_from_json: betterproto2.Message = plugin_module.Test().from_json(sample.json)
        plugin_instance_from_binary = plugin_module.Test.FromString(reference_binary_output)

        # Generally this can't be relied on, but here we are aiming to match the
        # existing Python implementation and aren't doing anything tricky.
        # https://developers.google.com/protocol-buffers/docs/encoding#implications
        assert bytes(plugin_instance_from_json) == reference_binary_output
        assert bytes(plugin_instance_from_binary) == reference_binary_output

        assert plugin_instance_from_json == plugin_instance_from_binary
        assert dict_replace_nans(plugin_instance_from_json.to_dict()) == dict_replace_nans(
            plugin_instance_from_binary.to_dict()
        )



def test_int_overflow():
    """Make sure that overflows in encoded values are handled correctly."""
    from tests.output_betterproto_pydantic.encoding_decoding import Overflow32, Overflow64

    b = bytes(Overflow64(uint=2**50 + 42))
    msg = Overflow32.parse(b)
    assert msg.uint == 42

    b = bytes(Overflow64(int=2**50 + 42))
    msg = Overflow32.parse(b)
    assert msg.int == 42

    b = bytes(Overflow64(int=2**50 - 42))
    msg = Overflow32.parse(b)
    assert msg.int == -42

    b = bytes(Overflow64(sint=2**50 + 42))
    msg = Overflow32.parse(b)
    assert msg.sint == 42

    b = bytes(Overflow64(sint=-(2**50) - 42))
    msg = Overflow32.parse(b)
    assert msg.sint == -42



import asyncio
import threading
from collections.abc import AsyncIterator

import grpc
import pytest
from grpclib.server import Server

from tests.output_betterproto.simple_service import Request, Response, SimpleServiceBase, SimpleServiceSyncStub


class SimpleService(SimpleServiceBase):
    async def get_unary_unary(self, message: "Request") -> "Response":
        return Response(message=f"Hello {message.value}")

    async def get_unary_stream(self, message: "Request") -> "AsyncIterator[Response]":
        for i in range(5):
            yield Response(message=f"Hello {message.value} {i}")

    async def get_stream_unary(self, messages: "AsyncIterator[Request]") -> "Response":
        s = 0
        async for m in messages:
            s += m.value
        return Response(message=f"Hello {s}")

    async def get_stream_stream(self, messages: "AsyncIterator[Request]") -> "AsyncIterator[Response]":
        async for message in messages:
            yield Response(message=f"Hello {message.value}")


@pytest.mark.asyncio
async def test_sync_client():
    start_server_event = threading.Event()
    close_server_event = asyncio.Event()

    def start_server():
        async def run_server():
            server = Server([SimpleService()])
            await server.start("127.0.0.1", 1234)
            start_server_event.set()

            await close_server_event.wait()
            server.close()

        loop = asyncio.new_event_loop()
        loop.run_until_complete(run_server())
        loop.close()

    # We need to start the server in a new thread to avoid a deadlock
    server_thread = threading.Thread(target=start_server)
    server_thread.start()

    # Create a sync client
    start_server_event.wait()

    with grpc.insecure_channel("localhost:1234") as channel:
        client = SimpleServiceSyncStub(channel)

        response = client.get_unary_unary(Request(value=42))
        assert response.message == "Hello 42"

        response = client.get_unary_stream(Request(value=42))
        assert [r.message for r in response] == [f"Hello 42 {i}" for i in range(5)]

        response = client.get_stream_unary([Request(value=i) for i in range(5)])
        assert response.message == "Hello 10"

        response = client.get_stream_stream([Request(value=i) for i in range(5)])
        assert [r.message for r in response] == [f"Hello {i}" for i in range(5)]

        close_server_event.set()

    # Create an async client
    # client = SimpleServiceStub(Channel(host="127.0.0.1", port=1234))
    # response = await client.get_unary_unary(Request(value=42))
    # assert response.message == "Hello 42"



import pydantic
import pytest


def test_int32_validation():
    from .output_betterproto_pydantic.validation import Message

    # Test valid values
    Message(int32_value=1)
    Message(int32_value=-(2**31))
    Message(int32_value=(2**31 - 1))

    # Test invalid values
    with pytest.raises(pydantic.ValidationError):
        Message(int32_value=2**31)
    with pytest.raises(pydantic.ValidationError):
        Message(int32_value=-(2**31) - 1)


def test_int64_validation():
    from .output_betterproto_pydantic.validation import Message

    # Test valid values
    Message(int64_value=1)
    Message(int64_value=-(2**63))
    Message(int64_value=(2**63 - 1))

    # Test invalid values
    with pytest.raises(pydantic.ValidationError):
        Message(int64_value=2**63)
    with pytest.raises(pydantic.ValidationError):
        Message(int64_value=-(2**63) - 1)


def test_uint32_validation():
    from .output_betterproto_pydantic.validation import Message

    # Test valid values
    Message(uint32_value=0)
    Message(uint32_value=2**32 - 1)

    # Test invalid values
    with pytest.raises(pydantic.ValidationError):
        Message(uint32_value=-1)
    with pytest.raises(pydantic.ValidationError):
        Message(uint32_value=2**32)


def test_uint64_validation():
    from .output_betterproto_pydantic.validation import Message

    # Test valid values
    Message(uint64_value=0)
    Message(uint64_value=2**64 - 1)

    # Test invalid values
    with pytest.raises(pydantic.ValidationError):
        Message(uint64_value=-1)
    with pytest.raises(pydantic.ValidationError):
        Message(uint64_value=2**64)


def test_sint32_validation():
    from .output_betterproto_pydantic.validation import Message

    # Test valid values
    Message(sint32_value=1)
    Message(sint32_value=-(2**31))
    Message(sint32_value=(2**31 - 1))

    # Test invalid values
    with pytest.raises(pydantic.ValidationError):
        Message(sint32_value=2**31)
    with pytest.raises(pydantic.ValidationError):
        Message(sint32_value=-(2**31) - 1)


def test_sint64_validation():
    from .output_betterproto_pydantic.validation import Message

    # Test valid values
    Message(sint64_value=1)
    Message(sint64_value=-(2**63))
    Message(sint64_value=(2**63 - 1))

    # Test invalid values
    with pytest.raises(pydantic.ValidationError):
        Message(sint64_value=2**63)
    with pytest.raises(pydantic.ValidationError):
        Message(sint64_value=-(2**63) - 1)


def test_fixed32_validation():
    from .output_betterproto_pydantic.validation import Message

    # Test valid values
    Message(fixed32_value=0)
    Message(fixed32_value=2**32 - 1)

    # Test invalid values
    with pytest.raises(pydantic.ValidationError):
        Message(fixed32_value=-1)
    with pytest.raises(pydantic.ValidationError):
        Message(fixed32_value=2**32)


def test_fixed64_validation():
    from .output_betterproto_pydantic.validation import Message

    # Test valid values
    Message(fixed64_value=0)
    Message(fixed64_value=2**64 - 1)

    # Test invalid values
    with pytest.raises(pydantic.ValidationError):
        Message(fixed64_value=-1)
    with pytest.raises(pydantic.ValidationError):
        Message(fixed64_value=2**64)


def test_sfixed32_validation():
    from .output_betterproto_pydantic.validation import Message

    # Test valid values
    Message(sfixed32_value=1)
    Message(sfixed32_value=-(2**31))
    Message(sfixed32_value=(2**31 - 1))

    # Test invalid values
    with pytest.raises(pydantic.ValidationError):
        Message(sfixed32_value=2**31)
    with pytest.raises(pydantic.ValidationError):
        Message(sfixed32_value=-(2**31) - 1)


def test_sfixed64_validation():
    from .output_betterproto_pydantic.validation import Message

    # Test valid values
    Message(sfixed64_value=1)
    Message(sfixed64_value=-(2**63))
    Message(sfixed64_value=(2**63 - 1))

    # Test invalid values
    with pytest.raises(pydantic.ValidationError):
        Message(sfixed64_value=2**63)
    with pytest.raises(pydantic.ValidationError):
        Message(sfixed64_value=-(2**63) - 1)


def test_float_validation():
    from .output_betterproto_pydantic.validation import Message

    # Test valid values
    Message(float_value=0.0)
    Message(float_value=3.14)

    # Test invalid values
    with pytest.raises(pydantic.ValidationError):
        Message(float_value=3.5e38)


def test_string_validation():
    from .output_betterproto_pydantic.validation import Message

    # Test valid values
    Message(string_value="")
    Message(string_value="Hello World")

    # Test invalid values
    with pytest.raises(pydantic.ValidationError):
        Message(string_value="Hello \udc00 World")



"""
Pydantic validators for Protocol Buffer standard types.

This module provides validator functions that can be used with Pydantic
to validate Protocol Buffer standard types (int32, int64, sfixed32, etc.)
to ensure they conform to their respective constraints.

These validators are designed to be used as "after validators", meaning the value
will already be of the correct type and only bounds checking is needed.
"""

import struct


def validate_float32(v: float) -> float:
    try:
        packed = struct.pack("!f", v)
        struct.unpack("!f", packed)
    except (struct.error, OverflowError):
        raise ValueError(f"Value cannot be encoded as a float: {v}")

    return v


def validate_string(v: str) -> str:
    try:
        v.encode("utf-8").decode("utf-8")
    except UnicodeError:
        raise ValueError("String contains invalid UTF-8 characters")
    return v



import betterproto2
from tests.output_betterproto.oneof_enum import Move, Signal, Test
from tests.util import get_test_case_json_data


def test_which_one_of_returns_enum_with_default_value():
    """
    returns first field when it is enum and set with default value
    """
    message = Test.from_json(get_test_case_json_data("oneof_enum", "oneof_enum-enum-0.json")[0].json)

    assert message.move is None
    assert message.signal == Signal.PASS
    assert betterproto2.which_one_of(message, "action") == ("signal", Signal.PASS)


def test_which_one_of_returns_enum_with_non_default_value():
    """
    returns first field when it is enum and set with non default value
    """
    message = Test.from_json(get_test_case_json_data("oneof_enum", "oneof_enum-enum-1.json")[0].json)

    assert message.move is None
    assert message.signal == Signal.RESIGN
    assert betterproto2.which_one_of(message, "action") == ("signal", Signal.RESIGN)


def test_which_one_of_returns_second_field_when_set():
    message = Test.from_json(get_test_case_json_data("oneof_enum")[0].json)
    assert message.move == Move(x=2, y=3)
    assert message.signal is None
    assert betterproto2.which_one_of(message, "action") == ("move", Move(x=2, y=3))



import json
from datetime import (
    datetime,
    timedelta,
    timezone,
)
from inspect import (
    Parameter,
    signature,
)
from unittest.mock import ANY

import betterproto2
from betterproto2 import OutputFormat


def test_class_init():
    from tests.output_betterproto.features import Bar, Foo

    foo = Foo(name="foo", child=Bar(name="bar"))

    assert foo.to_dict() == {"name": "foo", "child": {"name": "bar"}}
    assert foo.to_dict(output_format=OutputFormat.PYTHON) == {"name": "foo", "child": {"name": "bar"}}


def test_enum_as_int_json():
    from tests.output_betterproto.features import Enum, EnumMsg

    # JSON strings are supported, but ints should still be supported too.
    enum_msg = EnumMsg().from_dict({"enum": 1})
    assert enum_msg.enum == Enum.ONE

    # Plain-ol'-ints should serialize properly too.
    enum_msg.enum = 1
    assert enum_msg.to_dict() == {"enum": "ONE"}

    # Similar expectations for pydict
    enum_msg = EnumMsg().from_dict({"enum": 1})
    assert enum_msg.enum == Enum.ONE
    assert enum_msg.to_dict(output_format=OutputFormat.PYTHON) == {"enum": Enum.ONE}


def test_unknown_fields():
    from tests.output_betterproto.features import Newer, Older

    newer = Newer(x=True, y=1, z="Hello")
    serialized_newer = bytes(newer)

    # Unknown fields in `Newer` should round trip with `Older`
    round_trip = bytes(Older.parse(serialized_newer))
    assert serialized_newer == round_trip

    new_again = Newer.parse(round_trip)
    assert newer == new_again


def test_oneof_support():
    from tests.output_betterproto.features import IntMsg, OneofMsg

    msg = OneofMsg()

    assert betterproto2.which_one_of(msg, "group1")[0] == ""

    msg.x = 1
    assert betterproto2.which_one_of(msg, "group1")[0] == "x"

    msg.x = None
    msg.y = "test"
    assert betterproto2.which_one_of(msg, "group1")[0] == "y"

    msg.a = IntMsg(val=1)
    assert betterproto2.which_one_of(msg, "group2")[0] == "a"

    msg.a = None
    msg.b = "test"
    assert betterproto2.which_one_of(msg, "group2")[0] == "b"

    # Group 1 shouldn't be touched
    assert betterproto2.which_one_of(msg, "group1")[0] == "y"

    # Zero value should always serialize for one-of
    msg = OneofMsg(x=0)
    assert betterproto2.which_one_of(msg, "group1")[0] == "x"
    assert bytes(msg) == b"\x08\x00"

    # Round trip should also work
    msg = OneofMsg.parse(bytes(msg))
    assert betterproto2.which_one_of(msg, "group1")[0] == "x"
    assert msg.x == 0
    assert betterproto2.which_one_of(msg, "group2")[0] == ""


def test_json_casing():
    from tests.output_betterproto.features import JsonCasingMsg

    # Parsing should accept almost any input
    msg = JsonCasingMsg().from_dict({"PascalCase": 1, "camelCase": 2, "snake_case": 3, "kabob-case": 4})

    assert msg == JsonCasingMsg(1, 2, 3, 4)

    # Serializing should be strict.
    assert json.loads(msg.to_json()) == {
        "pascalCase": 1,
        "camelCase": 2,
        "snakeCase": 3,
        "kabobCase": 4,
    }

    assert json.loads(msg.to_json(casing=betterproto2.Casing.SNAKE)) == {
        "pascal_case": 1,
        "camel_case": 2,
        "snake_case": 3,
        "kabob_case": 4,
    }


def test_dict_casing():
    from tests.output_betterproto.features import JsonCasingMsg

    # Parsing should accept almost any input
    msg = JsonCasingMsg().from_dict({"PascalCase": 1, "camelCase": 2, "snake_case": 3, "kabob-case": 4})

    assert msg == JsonCasingMsg(1, 2, 3, 4)

    # Serializing should be strict.
    assert msg.to_dict() == {
        "pascalCase": 1,
        "camelCase": 2,
        "snakeCase": 3,
        "kabobCase": 4,
    }

    assert msg.to_dict(casing=betterproto2.Casing.SNAKE) == {
        "pascal_case": 1,
        "camel_case": 2,
        "snake_case": 3,
        "kabob_case": 4,
    }


def test_optional_flag():
    from tests.output_betterproto.features import OptionalBoolMsg

    # Serialization of not passed vs. set vs. zero-value.
    assert bytes(OptionalBoolMsg()) == b""
    assert bytes(OptionalBoolMsg(field=True)) == b"\n\x02\x08\x01"
    assert bytes(OptionalBoolMsg(field=False)) == b"\n\x00"

    # Differentiate between not passed and the zero-value.
    assert OptionalBoolMsg.parse(b"").field is None
    assert OptionalBoolMsg.parse(b"\n\x00").field is False


def test_optional_datetime_to_dict():
    from tests.output_betterproto.features import OptionalDatetimeMsg

    # Check dict serialization
    assert OptionalDatetimeMsg().to_dict() == {}
    assert OptionalDatetimeMsg().to_dict(include_default_values=True) == {"field": None}
    assert OptionalDatetimeMsg(field=datetime(2020, 1, 1, tzinfo=timezone.utc)).to_dict() == {
        "field": "2020-01-01T00:00:00Z"
    }
    assert OptionalDatetimeMsg(field=datetime(2020, 1, 1, tzinfo=timezone.utc)).to_dict(
        include_default_values=True
    ) == {"field": "2020-01-01T00:00:00Z"}

    # Check pydict serialization
    assert OptionalDatetimeMsg().to_dict(output_format=OutputFormat.PYTHON) == {}
    assert OptionalDatetimeMsg().to_dict(include_default_values=True, output_format=OutputFormat.PYTHON) == {
        "field": None
    }
    assert OptionalDatetimeMsg(field=datetime(2020, 1, 1)).to_dict(output_format=OutputFormat.PYTHON) == {
        "field": datetime(2020, 1, 1)
    }
    assert OptionalDatetimeMsg(field=datetime(2020, 1, 1)).to_dict(
        include_default_values=True, output_format=OutputFormat.PYTHON
    ) == {"field": datetime(2020, 1, 1)}


def test_to_json_default_values():
    from tests.output_betterproto.features import MsgA

    # Empty dict
    test = MsgA().from_dict({})

    assert json.loads(test.to_json(include_default_values=True)) == {
        "someInt": 0,
        "someDouble": 0.0,
        "someStr": "",
        "someBool": False,
    }

    # All default values
    test = MsgA().from_dict({"someInt": 0, "someDouble": 0.0, "someStr": "", "someBool": False})

    assert json.loads(test.to_json(include_default_values=True)) == {
        "someInt": 0,
        "someDouble": 0.0,
        "someStr": "",
        "someBool": False,
    }


def test_to_dict_default_values():
    from tests.output_betterproto.features import MsgA, MsgB

    # Empty dict
    test = MsgA()

    assert test.to_dict(include_default_values=True) == {
        "someInt": 0,
        "someDouble": 0.0,
        "someStr": "",
        "someBool": False,
    }

    assert test.to_dict(include_default_values=True, output_format=OutputFormat.PYTHON) == {
        "someInt": 0,
        "someDouble": 0.0,
        "someStr": "",
        "someBool": False,
    }

    # Some default and some other values
    test = MsgB().from_dict(
        {
            "someInt": 2,
            "someDouble": 1.2,
            "someStr": "hello",
            "someBool": True,
            "someDefaultInt": 0,
            "someDefaultDouble": 0.0,
            "someDefaultStr": "",
            "someDefaultBool": False,
        }
    )

    assert test.to_dict(include_default_values=True) == {
        "someInt": 2,
        "someDouble": 1.2,
        "someStr": "hello",
        "someBool": True,
        "someDefaultInt": 0,
        "someDefaultDouble": 0.0,
        "someDefaultStr": "",
        "someDefaultBool": False,
    }


def test_to_dict_datetime_values():
    from tests.output_betterproto.features import TimeMsg

    test = TimeMsg.from_dict({"timestamp": "2020-01-01T00:00:00Z", "duration": "86400s"})
    assert test.to_dict() == {"timestamp": "2020-01-01T00:00:00Z", "duration": "86400s"}

    test = TimeMsg.from_dict(
        {"timestamp": datetime(year=2020, month=1, day=1, tzinfo=timezone.utc), "duration": timedelta(days=1)}
    )
    assert test.to_dict(output_format=OutputFormat.PYTHON) == {
        "timestamp": datetime(year=2020, month=1, day=1, tzinfo=timezone.utc),
        "duration": timedelta(days=1),
    }
    assert test.to_dict(output_format=OutputFormat.PROTO_JSON) == {
        "timestamp": "2020-01-01T00:00:00Z",
        "duration": "86400s",
    }
    bytes(test)


def test_oneof_default_value_set_causes_writes_wire():
    from tests.output_betterproto.features import Empty, MsgC

    def _round_trip_serialization(msg: MsgC) -> MsgC:
        return MsgC.parse(bytes(msg))

    int_msg = MsgC(int_field=0)
    str_msg = MsgC(string_field="")
    empty_msg = MsgC(empty_field=Empty())
    msg = MsgC()

    assert bytes(int_msg) == b"\x08\x00"
    assert (
        betterproto2.which_one_of(int_msg, "group1")
        == betterproto2.which_one_of(_round_trip_serialization(int_msg), "group1")
        == ("int_field", 0)
    )

    assert bytes(str_msg) == b"\x12\x00"  # Baz is just an empty string
    assert (
        betterproto2.which_one_of(str_msg, "group1")
        == betterproto2.which_one_of(_round_trip_serialization(str_msg), "group1")
        == ("string_field", "")
    )

    assert bytes(empty_msg) == b"\x1a\x00"
    assert (
        betterproto2.which_one_of(empty_msg, "group1")
        == betterproto2.which_one_of(_round_trip_serialization(empty_msg), "group1")
        == ("empty_field", Empty())
    )

    assert bytes(msg) == b""
    assert (
        betterproto2.which_one_of(msg, "group1")
        == betterproto2.which_one_of(_round_trip_serialization(msg), "group1")
        == ("", None)
    )


def test_message_repr():
    from tests.output_betterproto.recursivemessage import Test

    assert repr(Test(name="Loki")) == "Test(name='Loki')"
    assert repr(Test(child=Test(), name="Loki")) == "Test(name='Loki', child=Test())"


def test_bool():
    """Messages should evaluate similarly to a collection
    >>> test = []
    >>> bool(test)
    ... False
    >>> test.append(1)
    >>> bool(test)
    ... True
    >>> del test[0]
    >>> bool(test)
    ... False
    """
    from tests.output_betterproto.features import Empty, IntMsg

    assert not Empty()
    t = IntMsg()
    assert not t
    t.val = 1
    assert t
    t.val = 0
    assert not t


# valid ISO datetimes according to https://www.myintervals.com/blog/2009/05/20/iso-8601-date-validation-that-doesnt-suck/
iso_candidates = """2009-12-12T12:34
2009
2009-05-19
2009-05-19
20090519
2009123
2009-05
2009-123
2009-222
2009-001
2009-W01-1
2009-W51-1
2009-W33
2009W511
2009-05-19
2009-05-19 00:00
2009-05-19 14
2009-05-19 14:31
2009-05-19 14:39:22
2009-05-19T14:39Z
2009-W21-2
2009-W21-2T01:22
2009-139
2009-05-19 14:39:22-06:00
2009-05-19 14:39:22+0600
2009-05-19 14:39:22-01
20090621T0545Z
2007-04-06T00:00
2007-04-05T24:00
2010-02-18T16:23:48.5
2010-02-18T16:23:48,444
2010-02-18T16:23:48,3-06:00
2010-02-18T16:23:00.4
2010-02-18T16:23:00,25
2010-02-18T16:23:00.33+0600
2010-02-18T16:00:00.23334444
2010-02-18T16:00:00,2283
2009-05-19 143922
2009-05-19 1439""".split("\n")


def test_iso_datetime():
    from tests.output_betterproto.features import TimeMsg

    for _, candidate in enumerate(iso_candidates):
        msg = TimeMsg.from_dict({"timestamp": candidate})
        assert isinstance(msg.timestamp, datetime)


def test_iso_datetime_list():
    from tests.output_betterproto.features import MsgD

    msg = MsgD()

    msg.from_dict({"timestamps": iso_candidates})
    assert all([isinstance(item, datetime) for item in msg.timestamps])


def test_service_argument__expected_parameter():
    from tests.output_betterproto.service import TestStub

    sig = signature(TestStub.do_thing)
    do_thing_request_parameter = sig.parameters["message"]
    assert do_thing_request_parameter.default is Parameter.empty
    assert do_thing_request_parameter.annotation == "DoThingRequest"


def test_is_set():
    from tests.output_betterproto.features import MsgE

    assert not MsgE().is_set("bool_field")
    assert not MsgE().is_set("int_field")
    assert MsgE(bool_field=True).is_set("bool_field")
    assert MsgE(bool_field=True, int_field=0).is_set("int_field")


def test_equality_comparison():
    from tests.output_betterproto.bool import Test as TestMessage

    msg = TestMessage(value=True)

    assert msg == msg
    assert msg == ANY
    assert msg == TestMessage(value=True)
    assert msg != 1
    assert msg != TestMessage(value=False)



import datetime


def test_message_wrapping_map():
    from tests.output_betterproto.message_wrapping import MapMessage

    msg = MapMessage(map1={"key": 12.0}, map2={"key": datetime.timedelta(seconds=1)})

    bytes(msg)

    assert msg.to_dict() == {"map1": {"key": 12.0}, "map2": {"key": "1s"}}



import importlib
import os
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType

os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"

root_path = Path(__file__).resolve().parent
inputs_path = root_path.joinpath("inputs")


def get_directories(path):
    for root, directories, files in os.walk(path):
        yield from directories


@dataclass
class TestCaseJsonFile:
    json: str
    test_name: str
    file_name: str

    def belongs_to(self, non_symmetrical_json: dict[str, tuple[str, ...]]) -> bool:
        return self.file_name in non_symmetrical_json.get(self.test_name, ())


def get_test_case_json_data(test_case_name: str, *json_file_names: str) -> list[TestCaseJsonFile]:
    """
    :return:
        A list of all files found in "{inputs_path}/test_case_name" with names matching
        f"{test_case_name}.json" or f"{test_case_name}_*.json", OR given by
        json_file_names
    """
    test_case_dir = inputs_path.joinpath(test_case_name)
    possible_file_paths = [
        *(test_case_dir.joinpath(json_file_name) for json_file_name in json_file_names),
        test_case_dir.joinpath(f"{test_case_name}.json"),
        *test_case_dir.glob(f"{test_case_name}_*.json"),
    ]

    result = []
    for test_data_file_path in possible_file_paths:
        if not test_data_file_path.exists():
            continue
        with test_data_file_path.open("r") as fh:
            result.append(TestCaseJsonFile(fh.read(), test_case_name, test_data_file_path.name.split(".")[0]))

    return result


def find_module(module: ModuleType, predicate: Callable[[ModuleType], bool]) -> ModuleType | None:
    """
    Recursively search module tree for a module that matches the search predicate.
    Assumes that the submodules are directories containing __init__.py.

    Example:

        # find module inside foo that contains Test
        import foo
        test_module = find_module(foo, lambda m: hasattr(m, 'Test'))
    """
    if predicate(module):
        return module

    module_path = Path(*module.__path__)

    for sub in [sub.parent for sub in module_path.glob("**/__init__.py")]:
        if sub == module_path:
            continue
        sub_module_path = sub.relative_to(module_path)
        sub_module_name = ".".join(sub_module_path.parts)

        sub_module = importlib.import_module(f".{sub_module_name}", module.__name__)

        if predicate(sub_module):
            return sub_module

    return None



import pickle
from copy import copy, deepcopy

import cachelib

from tests.output_betterproto.google import protobuf as google
from tests.output_betterproto.pickling import Complex, Fe, Fi, NestedData, PickledMessage


def unpickled(message):
    return pickle.loads(pickle.dumps(message))


def complex_msg():
    return Complex(
        foo_str="yep",
        fe=Fe(abc="1"),
        nested_data=NestedData(
            struct_foo={
                "foo": google.Struct(
                    fields={
                        "hello": google.Value(list_value=google.ListValue(values=[google.Value(string_value="world")]))
                    }
                ),
            },
            map_str_any_bar={
                "key": google.Any(value=b"value"),
            },
        ),
        mapping={
            "message": google.Any(value=bytes(Fi(abc="hi"))),
            "string": google.Any(value=b"howdy"),
        },
    )


def test_pickling_complex_message():
    msg = complex_msg()
    deser = unpickled(msg)
    assert msg == deser
    assert msg.fe.abc == "1"
    assert msg.is_set("fi") is not True
    assert msg.mapping["message"] == google.Any(value=bytes(Fi(abc="hi")))
    assert msg.mapping["string"].value.decode() == "howdy"
    assert msg.nested_data.struct_foo["foo"].fields["hello"].list_value.values[0].string_value == "world"


def test_recursive_message_defaults():
    from tests.output_betterproto.recursivemessage import Intermediate, Test as RecursiveMessage

    msg = RecursiveMessage(name="bob", intermediate=Intermediate(42))
    msg = unpickled(msg)

    # set values are as expected
    assert msg == RecursiveMessage(name="bob", intermediate=Intermediate(42))

    # lazy initialized works modifies the message
    assert msg != RecursiveMessage(name="bob", intermediate=Intermediate(42), child=RecursiveMessage(name="jude"))
    msg.child = RecursiveMessage(child=RecursiveMessage(name="jude"))
    assert msg == RecursiveMessage(
        name="bob",
        intermediate=Intermediate(42),
        child=RecursiveMessage(child=RecursiveMessage(name="jude")),
    )


def test_copyability():
    msg = PickledMessage(bar=12, baz=["hello"])
    msg = unpickled(msg)

    copied = copy(msg)
    assert msg == copied
    assert msg is not copied
    assert msg.baz is copied.baz

    deepcopied = deepcopy(msg)
    assert msg == deepcopied
    assert msg is not deepcopied
    assert msg.baz is not deepcopied.baz


def test_message_can_be_cached():
    """Cachelib uses pickling to cache values"""

    cache = cachelib.SimpleCache()

    def use_cache():
        calls = getattr(use_cache, "calls", 0)
        result = cache.get("message")
        if result is not None:
            return result
        else:
            setattr(use_cache, "calls", calls + 1)
            result = complex_msg()
            cache.set("message", result)
            return result

    for n in range(10):
        if n == 0:
            assert not cache.has("message")
        else:
            assert cache.has("message")

        msg = use_cache()
        assert use_cache.calls == 1  # The message is only ever built once
        assert msg.fe.abc == "1"
        assert msg.is_set("fi") is not True
        assert msg.mapping["message"] == google.Any(value=bytes(Fi(abc="hi")))
        assert msg.mapping["string"].value.decode() == "howdy"
        assert msg.nested_data.struct_foo["foo"].fields["hello"].list_value.values[0].string_value == "world"



def test_any() -> None:
    # TODO using a custom message pool will no longer be necessary when the well-known types will be compiled as well
    from tests.output_betterproto.any import Person
    from tests.output_betterproto.google.protobuf import Any

    person = Person(first_name="John", last_name="Smith")

    any = Any()
    any.pack(person)

    new_any = Any.parse(bytes(any))

    assert new_any.unpack() == person


def test_any_to_dict() -> None:
    from tests.output_betterproto.any import Person
    from tests.output_betterproto.google.protobuf import Any

    person = Person(first_name="John", last_name="Smith")

    any = Any()

    # TODO test with include defautl value
    assert any.to_dict() == {"@type": ""}

    # Pack an object inside
    any.pack(person)

    assert any.to_dict() == {
        "@type": "type.googleapis.com/any.Person",
        "firstName": "John",
        "lastName": "Smith",
    }

    # Pack again in another Any
    any2 = Any()
    any2.pack(any)

    assert any2.to_dict() == {
        "@type": "type.googleapis.com/google.protobuf.Any",
        "value": {"@type": "type.googleapis.com/any.Person", "firstName": "John", "lastName": "Smith"},
    }



import pytest


def test_oneof_pattern_matching():
    from tests.output_betterproto.features import IntMsg, OneofMsg

    msg = OneofMsg(y="test1", b="test2")

    match msg:
        case OneofMsg(x=int(_)):
            pytest.fail("Matched 'bar' instead of 'baz'")
        case OneofMsg(y=v):
            assert v == "test1"
        case _:
            pytest.fail("Matched neither 'bar' nor 'baz'")

    match msg:
        case OneofMsg(a=IntMsg(_)):
            pytest.fail("Matched 'sub' instead of 'abc'")
        case OneofMsg(b=v):
            assert v == "test2"
        case _:
            pytest.fail("Matched neither 'sub' nor 'abc'")

    msg.b = None
    msg.a = IntMsg(val=1)

    match msg:
        case OneofMsg(a=IntMsg(val=v)):
            assert v == 1
        case OneofMsg(b=str(v)):
            pytest.fail("Matched 'abc' instead of 'sub'")
        case _:
            pytest.fail("Matched neither 'sub' nor 'abc'")



import pytest


def test_unwrap() -> None:
    from betterproto2 import unwrap
    from tests.output_betterproto.unwrap import Message, NestedMessage

    with pytest.raises(ValueError):
        unwrap(Message().x)

    msg = Message(x=NestedMessage())
    assert msg.x == unwrap(msg.x)
